!!! Yi's Change (06/13/2014): 
!!!  1. define: bq, pwi
!!!  2. add from_param: w0 and  add from_param: pw1, pw2, pw3, pw4, for hamming window
!!!  3. add sep_read(bq)
!!!  4  call mem_alloc(bq, n1, n2)
!!!  5. call fbi_init4(w0, pw1, pw2, pw3, pw4)
!!!  6. call model_update2(vel) : add cp bq; cal vel2dt, gama, eta, tau, gama_mean
!!!  7. call fbi_alloc(wavelet)  
!!!  8. add from_param: qcomp 

!!! module change:
!!! fbi.f90
!!! prop.f90

program oper2

  use sep
  use mem_mod
  use fbi_mod

  implicit none

  integer                             :: nt, n1, n2, wfld_j, i
  integer                             :: sou_nx, sou_ox, sou_dx, sou_z, sou_x, isou
  integer                             :: rec_nx, rec_ox, rec_dx, rec_z, rec_x, irec
  integer                             :: bc1, bc2, bc3, bc4, nlag, olag, lag_type, geom_type
  real                                :: dt, d1, o1, d2, o2
  real				      :: w0  !!! Yi's change
  real				      :: pw1, pw2, pw3, pw4 !!! Yi's change
  logical                             :: qcomp  !!! Yi's change
  integer                             :: oper
  logical                             :: adj, sou_ghost, rec_ghost


  real, dimension(:), allocatable     :: wavelet
  real, dimension(:,:), allocatable   :: vel, deltaB
  real, dimension(:,:), allocatable   :: bq   !!! Yi's change
  real, dimension(:,:,:), allocatable :: refl, refl_mult, source, data, deltaD, deltaI

  integer, dimension(8)               :: timer0, timer1
  real                                :: timer_sum=0.
  real,dimension(1)		      ::myend

  call sep_init()

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer0)
!---------------------------------------------------------------------------

  call from_param("oper", oper)
  if(oper < 0 .or. oper > 4) call erexit("oper value is not correct")

  call from_history("n1", n1)
  call from_history("o1", o1)
  call from_history("d1", d1)
  call from_history("n2", n2)
  call from_history("o2", o2)
  call from_history("d2", d2)

  call from_param("lag_type", lag_type, 1) ! lag_type=1 is subsurface offset, lag_type=2 is time shift
  if(lag_type < 1 .or. lag_type > 2) call erexit("lag_type value is not correct")

  call from_param("geom_type", geom_type, 1) ! geom_type=1 is fixed receivers (land), geom_type=2 is fixed offset (marine)
  if(geom_type < 1 .or. geom_type > 2) call erexit("geom_type value is not correct")

  call from_param("nlag", nlag, 1)
  call from_param("olag", olag, 0)
  if(nlag<1-olag) write(0,*) "WARNING: zero lag is not included"

  call from_param("nt", nt)
  call from_aux("wavelet", "d1", dt)

  call from_param("bc1", bc1)
  call from_param("bc2", bc2)
  call from_param("bc3", bc3)
  call from_param("bc4", bc4)
  call from_param("wfld_j", wfld_j)
  if(bc1 < 0 .or. bc2 < 0 .or. bc3 < 0 .or. bc4 < 0) call erexit("bc values are not correct")
  if(wfld_j < 1) call erexit("wfld_j value is not correct")

  call from_param("sou_nx", sou_nx)
  call from_param("sou_ox", sou_ox)
  call from_param("sou_dx", sou_dx)
  call from_param("sou_z", sou_z)
  if(sou_ox < 1 .or. (sou_nx-1)*sou_dx+sou_ox > n2) call erexit("sou_ox value is not correct")
  if(sou_dx < 1) call erexit("sou_dx value is not correct")
  if(sou_z < 1 .or. sou_z > n1) call erexit("sou_z value is not correct")

  call from_param("rec_nx", rec_nx)
  call from_param("rec_ox", rec_ox)
  call from_param("rec_dx", rec_dx)
  call from_param("rec_z", rec_z)
  if(rec_dx < 1) call erexit("rec_dx value is not correct")
  if(rec_z < 1 .or. rec_z > n1) call erexit("rec_z value is not correct")

  if(geom_type .eq. 1) then
    if(rec_ox < 1 .or. (rec_nx-1)*rec_dx+rec_ox > n2) call erexit("rec axis values are not correct")
    if(rec_ox < 1+bc2 .or. (rec_nx-1)*rec_dx+rec_ox > n2-bc4) write(0,*) "WARNING: recording inside the absorbing boundaries"
  else if(geom_type .eq. 2) then
    if(rec_ox + sou_ox < 1 .or. (rec_nx-1)*rec_dx+rec_ox + (sou_nx-1)*sou_dx+sou_ox > n2) call erexit("rec axis values are not correct")
    if(rec_ox + sou_ox < 1+bc2 .or. (rec_nx-1)*rec_dx+rec_ox + (sou_nx-1)*sou_dx+sou_ox > n2-bc4) write(0,*) "WARNING: recording inside the absorbing boundaries"
    if(rec_nx<(0-rec_ox)/rec_dx+1) write(0,*) "WARNING: zero offset is not included"
  end if

  call from_param("adj", adj)
  call from_param("sou_ghost", sou_ghost)
  call from_param("rec_ghost", rec_ghost)

!!! Yi's change: adding Q param
  call from_param("qcomp", qcomp, .false.)
  call from_param("w0", w0, 1000.0)
  call from_param('pwz1', pw1, 0.6)
  call from_param('pwz2', pw2, 0.8)
  call from_param('pwx1', pw3, 0.6)
  call from_param('pwx2', pw4, 0.8)

  call fbi_init1(n1, o1, d1, n2, o2, d2, bc1, bc2, bc3, bc4)
  call fbi_init2(sou_nx, sou_ox, sou_dx, sou_z, rec_nx, rec_ox, rec_dx, rec_z)
  call fbi_init3(nt, dt, wfld_j, nlag, olag, lag_type, geom_type, sou_ghost, rec_ghost)
  call fbi_init4(w0, pw1, pw2, pw3, pw4)   !!! Yi's change

  if((oper .eq. 0) .or. (oper < 3 .and. .not. adj)) then
    call to_history("n1", nt)
    call to_history("o1", 0.)
    call to_history("d1", dt)
    call to_history("label1", "time")
    call to_history("n2", rec_nx)
    call to_history("o2", (rec_ox-1)*d2+o2)
    if(geom_type .eq. 2) call to_history("o2", rec_ox*d2)
    call to_history("d2", rec_dx*d2)
    call to_history("label2", "receiver location")
    if(geom_type .eq. 2) call to_history("label2", "receiver offset")
    call to_history("n3", sou_nx)
    call to_history("o3", (sou_ox-1)*d2+o2)
    call to_history("d3", sou_dx*d2)
    call to_history("label3", "source location")
  else
    call to_history("n1", n1)
    call to_history("o1", o1)
    call to_history("d1", d1)
    call to_history("label1", "depth")
    call to_history("n2", n2)
    call to_history("o2", o2)
    call to_history("d2", d2)
    call to_history("label2", "location")
    if((oper .eq. 1) .or. (oper .eq. 3 .and. .not. adj))then
      call to_history("n3", nlag)
      call to_history("o3", olag*d2)
      if(lag_type .eq. 2) call to_history("o3", olag*dt)
      call to_history("d3", d2)
      if(lag_type .eq. 2) call to_history("d3", dt)
      call to_history("label3", "subsurface offset")
      if(lag_type .eq. 2) call to_history("label3", "time lag")
    end if
  end if

! If the tag "wfldxx" is specified as a parameter, the history file will be created and the binary file will be written

  if(exist_file("wflds0")) then
    call to_history("n1", n1, "wflds0")
    call to_history("o1", o1, "wflds0")
    call to_history("d1", d1, "wflds0")
    call to_history("label1", "depth", "wflds0")
    call to_history("n2", n2, "wflds0")
    call to_history("o2", o2, "wflds0")
    call to_history("d2", d2, "wflds0")
    call to_history("label2", "location", "wflds0")
    call to_history("n3", nt, "wflds0")
    call to_history("o3", 0., "wflds0")
    call to_history("d3", dt, "wflds0")
    call to_history("label3", "time", "wflds0")
  endif

  if(exist_file("wflds1")) then
    call to_history("n1", n1, "wflds1")
    call to_history("o1", o1, "wflds1")
    call to_history("d1", d1, "wflds1")
    call to_history("label1", "depth", "wflds1")
    call to_history("n2", n2, "wflds1")
    call to_history("o2", o2, "wflds1")
    call to_history("d2", d2, "wflds1")
    call to_history("label2", "location", "wflds1")
    call to_history("n3", nt, "wflds1")
    call to_history("o3", 0., "wflds1")
    call to_history("d3", dt, "wflds1")
    call to_history("label3", "time", "wflds1")
  endif

  if(exist_file("wfldr0")) then
    call to_history("n1", n1, "wfldr0")
    call to_history("o1", o1, "wfldr0")
    call to_history("d1", d1, "wfldr0")
    call to_history("label1", "depth", "wfldr0")
    call to_history("n2", n2, "wfldr0")
    call to_history("o2", o2, "wfldr0")
    call to_history("d2", d2, "wfldr0")
    call to_history("label2", "location", "wfldr0")
    call to_history("n3", nt, "wfldr0")
    call to_history("o3", 0., "wfldr0")
    call to_history("d3", dt, "wfldr0")
    call to_history("label3", "time", "wfldr0")
  endif

  call mem_alloc(wavelet, nt)
  call mem_alloc(vel, n1, n2)
  call mem_alloc(bq, n1, n2)  !!! Yi's change

  call sep_read(vel)
  call sep_read(wavelet, "wavelet")
  call sep_read(bq,"bq")  !!! Yi's change

  call fbi_alloc(wavelet)
  call model_update2(vel, bq)

  if(oper .eq. 0) then ! Nonlinear operator: source <=> full nonlinear data (equation 9)
    call mem_alloc(source, nt, rec_nx, sou_nx)
    call mem_alloc(data, nt, rec_nx, sou_nx)
    if(adj) then
      do i=1,sou_nx
        call sep_read(data(:,:,i), "data")
      end do
      call L_nlop2(adj, .false.,qcomp, source, data)
      do i=1,sou_nx
        call sep_write(source(:,:,i))
      end do
    else
      do isou=1,sou_nx
        sou_x = (isou-1)*sou_dx + sou_ox
        if(geom_type .eq. 1) then
          irec = (sou_x-rec_ox)/rec_dx + 1
        else if(geom_type .eq. 2) then
          irec = (0-rec_ox)/rec_dx + 1
        end if
        source(:,irec,isou) = wavelet
      end do
      call L_nlop2(adj, .false.,qcomp, source, data)
      do i=1,sou_nx
        call sep_write(data(:,:,i))
      end do
    end if
  else if(oper .eq. 1) then ! Born operator: perturbation <=> First order scattered data (i.e. data due to perturbation) (equation 22)
    call mem_alloc(refl, n1, n2, nlag)
    call mem_alloc(data, nt, rec_nx, sou_nx)
    if(adj) then
      do i=1,sou_nx
        call sep_read(data(:,:,i), "data")
      end do
      call L_lop2(adj, .false.,qcomp, refl, data)
      do i=1,nlag
        call sep_write(refl(:,:,i))
      end do
    else
      do i=1,nlag
        call sep_read(refl(:,:,i), "refl")
      end do
      call L_lop2(adj, .false.,qcomp, refl, data)
      do i=1,sou_nx
        call sep_write(data(:,:,i))
      end do
    end if
  else if(oper .eq. 2) then ! Tomographic operator: background change <=> Second order scattered data (i.e. data due to transmission changes) (equation 25)
    call mem_alloc(refl, n1, n2, nlag)
    call mem_alloc(deltaB, n1, n2)
    call mem_alloc(deltaD, nt, rec_nx, sou_nx)
    do i=1,nlag
      call sep_read(refl(:,:,i), "refl")
    end do
    if(adj) then
      do i=1,sou_nx
        call sep_read(deltaD(:,:,i), "deltaD")
      end do
      call T_lop2(adj, .false.,qcomp, deltaB, deltaD, refl)
      call sep_write(deltaB)
    else
      call sep_read(deltaB, "deltaB")
      call T_lop2(adj, .false.,qcomp, deltaB, deltaD, refl)
      do i=1,sou_nx
        call sep_write(deltaD(:,:,i))
      end do
    end if
  else if(oper .eq. 3) then ! WEMVA operator: background change <=> Perturbation change (equation 29)
    call mem_alloc(data, nt, rec_nx, sou_nx)
    call mem_alloc(deltaB, n1, n2)
    call mem_alloc(deltaI, n1, n2, nlag)
    do i=1,sou_nx
      call sep_read(data(:,:,i), "data")
    end do
    if(adj) then
      do i=1,nlag
        call sep_read(deltaI(:,:,i), "deltaI")
      end do
      call W_lop2(adj, .false.,qcomp, deltaB, deltaI, data)
      call sep_write(deltaB)
    else
      call sep_read(deltaB, "deltaB")
      call W_lop2(adj, .false.,qcomp, deltaB, deltaI, data)
      do i=1,nlag
        call sep_write(deltaI(:,:,i))
      end do
    end if
  else if(oper .eq. 4) then ! Multiple scattering operator
    call mem_alloc(refl, n1, n2, nlag)
    call mem_alloc(refl_mult, n1, n2, nlag)
    call mem_alloc(data, nt, rec_nx, sou_nx)
    if(adj) then
      call erexit("Adjoint multiscattering is not implemented yet")
    else
      do i=1,nlag
        call sep_read(refl(:,:,i), "refl")
        call sep_read(refl_mult(:,:,i), "refl_mult")
      end do
      call M_lop2(adj, .false.,qcomp, refl, data, refl_mult)
      do i=1,sou_nx
        call sep_write(data(:,:,i))
      end do
    end if
  end if

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer1)
  timer0 = timer1-timer0
  timer_sum = timer_sum + (((timer0(3)*24+timer0(5))*60+timer0(6))*60+timer0(7))*1000+timer0(8)
  call to_history("timer sum (sec)", timer_sum/1000)
  call to_history("timer sum (min)", timer_sum/1000/60)
  write(0,*) "timer sum (sec)", timer_sum/1000
  write(0,*) "timer sum (min)", timer_sum/1000/60
  write(0,*) "DONE!"
  myend=1.0
  call sep_write(myend,"end")
!---------------------------------------------------------------------------

end program
