program Oper

  use sep
  use mem_mod
  use fbi_mod
  use nlcg_mod

  implicit none

  integer                             :: nt, n1, n2, wfld_j, i, stat
  integer                             :: sou_nx, sou_ox, sou_dx, sou_z, sou_x, isou
  integer                             :: rec_nx, rec_ox, rec_dx, rec_z, rec_x, irec
  integer                             :: bc1, bc2, bc3, bc4, geom_type
  real                                :: dt, d1, o1, d2, o2

  logical                             :: sou_ghost, rec_ghost

  real, dimension(:), allocatable     :: wavelet, res
  real, dimension(:), allocatable     :: vel, slowsq, gradient, vmin, vmax, ssmin, ssmax
  real, dimension(:), allocatable     :: source, data_obs, data

  integer                             :: task, iter, niter, ntry
  double precision                    :: f

  integer, dimension(8)               :: timer0, timer1
  real                                :: timer_sum=0.

  call sep_init()

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer0)
!---------------------------------------------------------------------------

  call from_history("n1", n1)
  call from_history("o1", o1)
  call from_history("d1", d1)
  call from_history("n2", n2)
  call from_history("o2", o2)
  call from_history("d2", d2)

  call from_param("geom_type", geom_type, 1) ! geom_type=1 is fixed receivers (land), geom_type=2 is fixed offset (marine)
  if(geom_type < 1 .or. geom_type > 2) call erexit("geom_type value is not correct")

  call from_param("nt", nt)
  call from_aux("wavelet", "d1", dt)

  call from_param("bc1", bc1)
  call from_param("bc2", bc2)
  call from_param("bc3", bc3)
  call from_param("bc4", bc4)
  call from_param("wfld_j", wfld_j)
  if(bc1 < 0 .or. bc2 < 0 .or. bc3 < 0 .or. bc4 < 0) call erexit("bc values are not correct")
  if(wfld_j < 1) call erexit("wfld_j value is not correct")

  call from_param("sou_nx", sou_nx)
  call from_param("sou_ox", sou_ox)
  call from_param("sou_dx", sou_dx)
  call from_param("sou_z", sou_z)
  if(sou_ox < 1 .or. (sou_nx-1)*sou_dx+sou_ox > n2) call erexit("sou_ox value is not correct")
  if(sou_dx < 1) call erexit("sou_dx value is not correct")
  if(sou_z < 1 .or. sou_z > n1) call erexit("sou_z value is not correct")

  call from_param("rec_nx", rec_nx)
  call from_param("rec_ox", rec_ox)
  call from_param("rec_dx", rec_dx)
  call from_param("rec_z", rec_z)
  if(rec_dx < 1) call erexit("rec_dx value is not correct")
  if(rec_z < 1 .or. rec_z > n1) call erexit("rec_z value is not correct")

  if(geom_type .eq. 1) then
    if(rec_ox < 1 .or. (rec_nx-1)*rec_dx+rec_ox > n2) call erexit("rec axis values are not correct")
    if(rec_ox < 1+bc2 .or. (rec_nx-1)*rec_dx+rec_ox > n2-bc4) write(0,*) "WARNING: recording inside the absorbing boundaries"
  else if(geom_type .eq. 2) then
    if(rec_ox + sou_ox < 1 .or. (rec_nx-1)*rec_dx+rec_ox + (sou_nx-1)*sou_dx+sou_ox > n2) call erexit("rec axis values are not correct")
    if(rec_ox + sou_ox < 1+bc2 .or. (rec_nx-1)*rec_dx+rec_ox + (sou_nx-1)*sou_dx+sou_ox > n2-bc4) write(0,*) "WARNING: recording inside the absorbing boundaries"
    if(rec_nx<(0-rec_ox)/rec_dx+1) write(0,*) "WARNING: zero offset is not included"
  end if

  call from_param("sou_ghost", sou_ghost)
  call from_param("rec_ghost", rec_ghost)

  call from_param('niter', niter, 10)
  call from_param('ntry', ntry, 5)

  call fbi_init1(n1, o1, d1, n2, o2, d2, bc1, bc2, bc3, bc4)
  call fbi_init2(sou_nx, sou_ox, sou_dx, sou_z, rec_nx, rec_ox, rec_dx, rec_z)
  call fbi_init3(nt, dt, wfld_j, 1, 0, 1, geom_type, sou_ghost, rec_ghost)

  call to_history("n1", n1)
  call to_history("o1", o1)
  call to_history("d1", d1)
  call to_history("label1", "depth")
  call to_history("n2", n2)
  call to_history("o2", o2)
  call to_history("d2", d2)
  call to_history("label2", "location")

  call to_history("n1", n1, "mmov")
  call to_history("o1", o1, "mmov")
  call to_history("d1", d1, "mmov")
  call to_history("label1", "depth", "mmov")
  call to_history("n2", n2, "mmov")
  call to_history("o2", o2, "mmov")
  call to_history("d2", d2, "mmov")
  call to_history("label2", "location", "mmov")
  call to_history("o3", 0., "mmov")
  call to_history("d3", 1., "mmov")
  call to_history("label3", "iteration", "mmov")

  call to_history("n1", 2, "res")
  call to_history("o1", 0., "res")
  call to_history("d1", 1., "res")
  call to_history("label1", "component", "res")
  call to_history("o2", 0., "res")
  call to_history("d2", 1., "res")
  call to_history("label2", "iteration", "res")

  call mem_alloc(wavelet, nt)
  call mem_alloc(vel, n1*n2)
  call mem_alloc(res, 2)
  call mem_alloc(slowsq, n1*n2)
  call mem_alloc(gradient, n1*n2)
  call mem_alloc(vmin, n1*n2)
  call mem_alloc(vmax, n1*n2)
  call mem_alloc(ssmin, n1*n2)
  call mem_alloc(ssmax, n1*n2)
  call mem_alloc(source, nt*rec_nx*sou_nx)
  call mem_alloc(data, nt*rec_nx*sou_nx)
  call mem_alloc(data_obs, nt*rec_nx*sou_nx)

  call sep_read(vel)
  call sep_read(wavelet, "wavelet")
  do isou=1,sou_nx
    call sep_read(data_obs(1+(isou-1)*nt*rec_nx:isou*nt*rec_nx), "data")
  end do

  call sep_read(vmin, "vmin")
  call sep_read(vmax, "vmax")

  do isou=1,sou_nx
    sou_x = (isou-1)*sou_dx + sou_ox
    if(geom_type .eq. 1) then
      irec = (sou_x-rec_ox)/rec_dx + 1
    else if(geom_type .eq. 2) then
      irec = (0-rec_ox)/rec_dx + 1
    end if
    source(1+nt*(irec-1+(isou-1)*rec_nx):nt*(irec+(isou-1)*rec_nx)) = wavelet
  end do

  ssmax = 1./(vmin*vmin)
  ssmin = 1./(vmax*vmax)

  slowsq = 1./(vel*vel)

  call fbi_alloc(wavelet)
  call nlcg_init(n1*n2, ntry, ssmin, ssmax)

  task = 0
  iter = 1

  do while(iter <= niter .and. task < 6)
    call nlcg(f, gradient, slowsq, task)
    if(task .eq. 1) then
      vel = 1./sqrt(slowsq)
      stat = model_update(vel)
      stat = L_nlop(.false., .false., source, data)
      data = data - data_obs
      f = sum(dprod(data,data))
      stat = L_lop(.true., .false., gradient, data)
      res(1) = real(f)
      res(2) = 0.
    else if(task > 1 .and. task < 5) then
      vel = 1./sqrt(slowsq)
      stat = model_update(vel)
      stat = L_nlop(.false., .false., source, data)
      data = data - data_obs
      f = sum(dprod(data,data))
    else if(task .eq. 5) then
      ! Output model
      vel = 1./sqrt(slowsq)
      call sep_write(vel, "mmov")
      call sep_write(res, "res")
      call to_history("n3", iter, "mmov")
      call to_history("n2", iter, "res")
      iter = iter + 1
    end if
  end do

  call sep_write(vel)

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer1)
  timer0 = timer1-timer0
  timer_sum = timer_sum + (((timer0(3)*24+timer0(5))*60+timer0(6))*60+timer0(7))*1000+timer0(8)
!---------------------------------------------------------------------------
!---------------------------------------------------------------------------
  call to_history("timer sum (sec)", timer_sum/1000)
  call to_history("timer sum (min)", timer_sum/1000/60)
  write(0,*) "timer sum (sec)", timer_sum/1000
  write(0,*) "timer sum (min)", timer_sum/1000/60
  write(0,*) "DONE!"
!---------------------------------------------------------------------------

  call sep_close()

end program
