module escat_refl_mod

  implicit none

  integer, private                         :: n1, n2, nt, nlag, olag, lag_type, it, ilag, lag, ix, iz

  contains

  subroutine escat_refl_init(n1_in, n2_in, nt_in, nlag_in, olag_in, lag_type_in)
    integer                        :: n1_in, n2_in, nt_in, nlag_in, olag_in, lag_type_in
    n1 = n1_in
    n2 = n2_in
    nt = nt_in
    nlag = nlag_in
    olag = olag_in
    lag_type = lag_type_in
  end subroutine

  function escat_refl_op(adj, add, model, data, refl) result(stat)
    optional           :: refl
    logical,intent(in) :: adj, add
    real,dimension(:)  :: model, data, refl
    integer            :: stat
    if(.not. present(refl)) call erexit("optional parameter refl is required")
    call escat_refl_op2(adj, add, model, data, refl)
    stat=0
  end function

  subroutine escat_refl_op2(adj, add, model, data, refl)
    optional                    :: refl
    logical,intent(in)          :: adj, add 
    real, dimension(n1,n2,nt)   :: model, data
    real, dimension(n1,n2,nlag) :: refl
    if(.not. present(refl)) call erexit("optional parameter refl is required")
    if(adj) then
      if(.not. add) model = 0.
      if(lag_type .eq. 1) then ! Subsurface offset
        do it=1,nt
          do ilag=1,nlag
            lag = ilag - 1 + olag
            do ix=1+abs(lag),n2-abs(lag)
              model(:,ix-lag,it) = model(:,ix-lag,it) + refl(:,ix,ilag)*data(:,ix+lag,it)
            end do
          end do
        end do
      else if(lag_type .eq. 2) then ! Time shifts
        do ilag=1,nlag
          lag = ilag - 1 + olag
          do it=1+abs(lag),nt-abs(lag)
            model(:,:,it-lag) = model(:,:,it-lag) + refl(:,:,ilag)*data(:,:,it+lag)
          end do
        end do
      end if
    else
      if(.not. add) data = 0.
      if(lag_type .eq. 1) then ! Subsurface offset
        do it=1,nt
          do ilag=1,nlag
            lag = ilag - 1 + olag
            do ix=1+abs(lag),n2-abs(lag)
              data(:,ix+lag,it) = data(:,ix+lag,it) + refl(:,ix,ilag)*model(:,ix-lag,it)
            end do
          end do
        end do
      else if(lag_type .eq. 2) then ! Time shifts
        do ilag=1,nlag
          lag = ilag - 1 + olag
          do it=1+abs(lag),nt-abs(lag)
            data(:,:,it+lag) = data(:,:,it+lag) + refl(:,:,ilag)*model(:,:,it-lag)
          end do
        end do
      end if
    end if
  end subroutine

end module
