module escats_mod

  implicit none

  integer, private                         :: n1, n2, nt, nlag, olag, lag_type, it, ilag, lag, ix, iz

  contains

  subroutine escats_init(n1_in, n2_in, nt_in, nlag_in, olag_in, lag_type_in)
    integer                        :: n1_in, n2_in, nt_in, nlag_in, olag_in, lag_type_in
    n1 = n1_in
    n2 = n2_in
    nt = nt_in
    nlag = nlag_in
    olag = olag_in
    lag_type = lag_type_in
  end subroutine

  function escats_op(adj, add, model, data, wflds) result(stat)
    optional           :: wflds
    logical,intent(in) :: adj, add
    real,dimension(:)  :: model, data, wflds
    integer            :: stat
    if(.not. present(wflds)) call erexit("optional parameter wflds is required")
    call escats_op2(adj, add, model, data, wflds)
    stat=0
  end function

  subroutine escats_op2(adj, add, model, data, wflds)
    optional                    :: wflds
    logical,intent(in)          :: adj, add 
    real, dimension(n1,n2,nlag) :: model
    real, dimension(n1,n2,nt)   :: data, wflds
    if(.not. present(wflds)) call erexit("optional parameter wflds is required")
    if(adj) then
      if(.not. add) model = 0.
      if(lag_type .eq. 1) then ! Subsurface offset
        do it=1,nt
          do ilag=1,nlag
            lag = ilag - 1 + olag
            do ix=1+abs(lag),n2-abs(lag)
              model(:,ix,ilag) = model(:,ix,ilag) + data(:,ix+lag,it)*wflds(:,ix-lag,it)
            end do
          end do
        end do
      else if(lag_type .eq. 2) then ! Time shifts
        do ilag=1,nlag
          lag = ilag - 1 + olag
          do it=1+abs(lag),nt-abs(lag)
            model(:,:,ilag) = model(:,:,ilag) + data(:,:,it+lag)*wflds(:,:,it-lag)
          end do
        end do
      end if
    else
      if(.not. add) data = 0.
      if(lag_type .eq. 1) then ! Subsurface offset
        do it=1,nt
          do ilag=1,nlag
            lag = ilag - 1 + olag
            do ix=1+abs(lag),n2-abs(lag)
              data(:,ix+lag,it) = data(:,ix+lag,it) + model(:,ix,ilag)*wflds(:,ix-lag,it)
            end do
          end do
        end do
      else if(lag_type .eq. 2) then ! Time shifts
        do ilag=1,nlag
          lag = ilag - 1 + olag
          do it=1+abs(lag),nt-abs(lag)
            data(:,:,it+lag) = data(:,:,it+lag) + model(:,:,ilag)*wflds(:,:,it-lag)
          end do
        end do
      end if
    end if
  end subroutine

end module
