module prop_mod

  use mem_mod

  implicit none

  integer, private                            :: n1, n2, nt, wfld_j
  integer, private                            :: bc1, bc2, bc3, bc4, fat
  integer, private                            :: iz, ix, it, iw
  real, private                               :: stencil0, c
  real, dimension(5), private                 :: stencilz, stencilx
  real, dimension(:), allocatable, private    :: lwt1, lwt2, lwt3, lwt4
  real, dimension(:,:), pointer, private      :: veldt2, vwt, wt
  real, dimension(:,:), allocatable, private  :: p0, p1, p2

  contains

  subroutine prop_init(n1_in, n2_in, nt_in, wfld_j_in, bc1_in, bc2_in, bc3_in, bc4_in, d1_in, d2_in, veldt2_in, vwt_in, wt_in)
    integer                        :: n1_in, n2_in, nt_in, wfld_j_in, bc1_in, bc2_in, bc3_in, bc4_in
    real                           :: d1_in, d2_in
    real, dimension(:,:), target   :: veldt2_in, vwt_in, wt_in

    n1 = n1_in
    n2 = n2_in
    nt = nt_in
    wfld_j = wfld_j_in
    bc1 = bc1_in
    bc2 = bc2_in
    bc3 = bc3_in
    bc4 = bc4_in

    veldt2 => veldt2_in
    vwt => vwt_in
    wt => wt_in

    fat = 5

    call mem_alloc(lwt1, wfld_j)
    call mem_alloc(lwt2, wfld_j)
    call mem_alloc(lwt3, wfld_j)
    call mem_alloc(lwt4, wfld_j)
    do iw=1,wfld_j
      lwt2(iw) = 1.*iw/wfld_j
      lwt4(iw) = 1.*(iw-1)/wfld_j
    end do
    lwt1 = 1. - lwt2
    lwt3 = 1. - lwt4
    c = sqrt(1.*wfld_j)
    lwt1 = lwt1 / c
    lwt2 = lwt2 / c
    lwt3 = lwt3 / c
    lwt4 = lwt4 / c

    stencilz(1)=3.333333333/(2.*d1_in*d1_in)
    stencilz(2)=-0.4761904762/(2.*d1_in*d1_in)
    stencilz(3)=0.0793650794/(2.*d1_in*d1_in)
    stencilz(4)=-0.0099206349/(2.*d1_in*d1_in)
    stencilz(5)=0.0006349206/(2.*d1_in*d1_in)

    stencilx(1)=3.333333333/(2.*d2_in*d2_in)
    stencilx(2)=-0.4761904762/(2.*d2_in*d2_in)
    stencilx(3)=0.0793650794/(2.*d2_in*d2_in)
    stencilx(4)=-0.0099206349/(2.*d2_in*d2_in)
    stencilx(5)=0.0006349206/(2.*d2_in*d2_in)

    stencil0=(stencilz(5)+stencilx(5))
    stencil0=stencil0+(stencilz(4)+stencilx(4))
    stencil0=stencil0+(stencilz(3)+stencilx(3))
    stencil0=stencil0+(stencilz(2)+stencilx(2))
    stencil0=stencil0+(stencilz(1)+stencilx(1))
    stencil0=-2.*stencil0

    if(.not. allocated(p0)) then
      allocate(p0(1-fat:n1+fat,1-fat:n2+fat))
      call mem_add_4(size(p0))
    end if
    if(.not. allocated(p1)) then
      allocate(p1(1-fat:n1+fat,1-fat:n2+fat))
      call mem_add_4(size(p1))
    end if
    if(.not. allocated(p2)) then
      allocate(p2(1-fat:n1+fat,1-fat:n2+fat))
      call mem_add_4(size(p2))
    end if

  end subroutine

  function prop_op(adj, add, model, data) result(stat)
    logical,intent(in) :: adj, add
    real,dimension(:)  :: model, data
    integer            :: stat
    call prop_op2(adj, add, model, data)
    stat=0
  end function

  subroutine prop_op2(adj, add, model, data)
    logical,intent(in)        :: adj, add 
    real, dimension(n1,n2,nt) :: model, data

    p0 = 0.
    p1 = 0.
    p2 = 0.

    if(adj) then
      if(.not. add) model = 0.
      do it=nt,2,-1
        do iw=1,wfld_j
          p2(1:n1,1:n2) = lwt3(iw)*data(:,:,it) + lwt4(iw)*data(:,:,it-1) + (2.-wt)*p1(1:n1,1:n2) + (wt-1.)*p0(1:n1,1:n2)
          p0(1:n1,1:n2) = veldt2*(1.-wt)*p1(1:n1,1:n2)
          do ix=1,n2
            do iz=0,1-fat,-1
              p0(iz,ix) = veldt2(1,ix)*(1.-wt(1,ix))*p1(iz,ix)
            end do
            do iz=n1+1,n1+fat
              p0(iz,ix) = veldt2(n1,ix)*(1.-wt(n1,ix))*p1(iz,ix)
            end do
          end do
          do ix=0,1-fat,-1
            do iz=1,n1
              p0(iz,ix) = veldt2(iz,1)*(1.-wt(iz,1))*p1(iz,ix)
            end do
          end do
          do ix=n2+1,n2+fat
            do iz=1,n1
              p0(iz,ix) = veldt2(iz,n2)*(1.-wt(iz,n2))*p1(iz,ix)
            end do
          end do
          do ix=1,n2
            do iz=1,n1
              p2(iz,ix) = p2(iz,ix) +&
                          stencil0*p0(iz,ix)+&
                          stencilz(1)*(p0(iz-1,ix  )+p0(iz+1,ix  ))+&
                          stencilx(1)*(p0(iz  ,ix-1)+p0(iz  ,ix+1))+&
                          stencilz(2)*(p0(iz-2,ix  )+p0(iz+2,ix  ))+&
                          stencilx(2)*(p0(iz  ,ix-2)+p0(iz  ,ix+2))+&
                          stencilz(3)*(p0(iz-3,ix  )+p0(iz+3,ix  ))+&
                          stencilx(3)*(p0(iz  ,ix-3)+p0(iz  ,ix+3))+&
                          stencilz(4)*(p0(iz-4,ix  )+p0(iz+4,ix  ))+&
                          stencilx(4)*(p0(iz  ,ix-4)+p0(iz  ,ix+4))+&
                          stencilz(5)*(p0(iz-5,ix  )+p0(iz+5,ix  ))+&
                          stencilx(5)*(p0(iz  ,ix-5)+p0(iz  ,ix+5))
            end do
          end do

          p0(1:n1,1:n2) = wt*vwt*p1(1:n1,1:n2)
          do ix=1,n2
            do iz=1,bc1
              p2(iz,ix) = p2(iz,ix) - p0(iz,ix)
              p2(iz+1,ix) = p2(iz+1,ix) + p0(iz,ix)
            end do
            do iz=n1-bc3+1,n1
              p2(iz,ix) = p2(iz,ix) - p0(iz,ix)
              p2(iz-1,ix) = p2(iz-1,ix) + p0(iz,ix)
            end do
          end do
          do ix=1,bc2
            do iz=1,n1
              p2(iz,ix) = p2(iz,ix) - p0(iz,ix)
              p2(iz,ix+1) = p2(iz,ix+1) + p0(iz,ix)
            end do
          end do
          do ix=n2-bc4+1,n2
            do iz=1,n1
              p2(iz,ix) = p2(iz,ix) - p0(iz,ix)
              p2(iz,ix-1) = p2(iz,ix-1) + p0(iz,ix)
            end do
          end do

          do ix=1,n2
            do iz=0,1-fat,-1
              p2(iz,ix) = (1.-1.2*vwt(1,ix))/(1.+1.2*vwt(1,ix)) * (p1(iz,ix)-p2(iz+1,ix)) + p1(iz+1,ix)
            end do
            do iz=n1+1,n1+fat
              p2(iz,ix) = (1.-1.2*vwt(n1,ix))/(1.+1.2*vwt(n1,ix)) * (p1(iz,ix)-p2(iz-1,ix)) + p1(iz-1,ix)
            end do
          end do
          do ix=0,1-fat,-1
            do iz=1,n1
              p2(iz,ix) = (1.-1.2*vwt(iz,1))/(1.+1.2*vwt(iz,1)) * (p1(iz,ix)-p2(iz,ix+1)) + p1(iz,ix+1)
            end do
          end do
          do ix=n2+1,n2+fat
            do iz=1,n1
              p2(iz,ix) = (1.-1.2*vwt(iz,n2))/(1.+1.2*vwt(iz,n2)) * (p1(iz,ix)-p2(iz,ix-1)) + p1(iz,ix-1)
            end do
          end do

          model(:,:,it) = model(:,:,it) + lwt1(iw)*p2(1:n1,1:n2)
          model(:,:,it-1) = model(:,:,it-1) + lwt2(iw)*p2(1:n1,1:n2)
          p0 = p1
          p1 = p2
        end do
      end do
    else
      if(.not. add) data = 0.
      do it=1,nt-1
        do iw=1,wfld_j
   	  !!! Yi: p(it)=c4f(it)???+[(I-w)*2I+w*I]*p(it-1)+(w-I)*p(it-2)
          p2(1:n1,1:n2) = lwt3(iw)*model(:,:,it) + lwt4(iw)*model(:,:,it+1) + (2.-wt)*p1(1:n1,1:n2) + (wt-1.)*p0(1:n1,1:n2)
	!!! Yi:v^2*dt^2*Laplacian*(1-wt)
          do ix=1,n2
            do iz=1,n1
              p2(iz,ix) = p2(iz,ix) +&
                          veldt2(iz,ix)*(1.-wt(iz,ix))*(stencil0*p1(iz,ix)+&
                          stencilz(1)*(p1(iz-1,ix  )+p1(iz+1,ix  ))+&
                          stencilx(1)*(p1(iz  ,ix-1)+p1(iz  ,ix+1))+&
                          stencilz(2)*(p1(iz-2,ix  )+p1(iz+2,ix  ))+&
                          stencilx(2)*(p1(iz  ,ix-2)+p1(iz  ,ix+2))+&
                          stencilz(3)*(p1(iz-3,ix  )+p1(iz+3,ix  ))+&
                          stencilx(3)*(p1(iz  ,ix-3)+p1(iz  ,ix+3))+&
                          stencilz(4)*(p1(iz-4,ix  )+p1(iz+4,ix  ))+&
                          stencilx(4)*(p1(iz  ,ix-4)+p1(iz  ,ix+4))+&
                          stencilz(5)*(p1(iz-5,ix  )+p1(iz+5,ix  ))+&
                          stencilx(5)*(p1(iz  ,ix-5)+p1(iz  ,ix+5)))
            end do
          end do 

          do ix=1,n2
            do iz=1,bc1
              p2(iz,ix) = p2(iz,ix) - wt(iz,ix)*vwt(iz,ix)*(p1(iz,ix)-p1(iz+1,ix))
            end do
            do iz=n1-bc3+1,n1
              p2(iz,ix) = p2(iz,ix) - wt(iz,ix)*vwt(iz,ix)*(p1(iz,ix)-p1(iz-1,ix))
            end do
          end do
          do ix=1,bc2
            do iz=1,n1
              p2(iz,ix) = p2(iz,ix) - wt(iz,ix)*vwt(iz,ix)*(p1(iz,ix)-p1(iz,ix+1))
            end do
          end do
          do ix=n2-bc4+1,n2
            do iz=1,n1
              p2(iz,ix) = p2(iz,ix) - wt(iz,ix)*vwt(iz,ix)*(p1(iz,ix)-p1(iz,ix-1))
            end do
          end do

          do ix=1,n2
            do iz=0,1-fat,-1
              p2(iz,ix) = (1.-1.2*vwt(1,ix))/(1.+1.2*vwt(1,ix)) * (p1(iz,ix)-p2(iz+1,ix)) + p1(iz+1,ix)
            end do
            do iz=n1+1,n1+fat
              p2(iz,ix) = (1.-1.2*vwt(n1,ix))/(1.+1.2*vwt(n1,ix)) * (p1(iz,ix)-p2(iz-1,ix)) + p1(iz-1,ix)
            end do
          end do
          do ix=0,1-fat,-1
            do iz=1,n1
              p2(iz,ix) = (1.-1.2*vwt(iz,1))/(1.+1.2*vwt(iz,1)) * (p1(iz,ix)-p2(iz,ix+1)) + p1(iz,ix+1)
            end do
          end do
          do ix=n2+1,n2+fat
            do iz=1,n1
              p2(iz,ix) = (1.-1.2*vwt(iz,n2))/(1.+1.2*vwt(iz,n2)) * (p1(iz,ix)-p2(iz,ix-1)) + p1(iz,ix-1)
            end do
          end do

          data(:,:,it) = data(:,:,it) + lwt1(iw)*p2(1:n1,1:n2)
          data(:,:,it+1) = data(:,:,it+1) + lwt2(iw)*p2(1:n1,1:n2)
          p0 = p1
          p1 = p2
        end do
      end do
    end if
  end subroutine

end module
