module prop_base
  implicit none

contains
  !!----------------------------------------------------------
  !!.. Calculate the k map
  !!----------------------------------------------------------
  subroutine k_map(k,n,d)
    real,  dimension(:) :: k
    integer, intent(in) :: n
    real,    intent(in) :: d
    integer             :: i,m
    real, parameter     :: PI = 3.14159265359
    k=0.
    if(n/=1) then
       m = n/2
       do i=1,m+1
          k(i)=i-1
       end do
       do i=m+2,n
          k(i)=i-n-1
       end do
    else
       k=0. !! make sure k=0 if n=1
    end if

    k = k / (n*d)      !! 1/length
    k = 2 * PI * k     !! radians
    return
  end subroutine
  !!----------------------------------------------------------
  !!.. Calculate the k map with high frequency filtered
  !!----------------------------------------------------------
  subroutine k_map2(k,n,d,a)
    real,  dimension(:) :: k
    integer, intent(in) :: n
    real,    intent(in) :: d
    integer             :: i,m,a
    real, parameter     :: PI = 3.14159265359
    k=0.
    if(n/=1) then
       m = n/2
       do i=1,m+1-a
          k(i)=i-1
       end do
       do i=m+2+a,n
          k(i)=i-n-1
       end do
    else
       k=0. !! make sure k=0 if n=1
    end if

    k = k / (n*d)      !! 1/length
    k = 2 * PI * k     !! radians
    return
  end subroutine
  !----------------------------------------------
  ! Hamming window taper
  !----------------------------------------------
  subroutine hammingw(ctmp, pw1, pw2, pw3, pw4)
    complex, dimension(:)         :: ctmp
    real            :: pw1, pw2, pw3, pw4
    integer         :: nw1, nw2, nw3, nw4
    integer         :: iw, nt
    real(kind=4)    :: s
    real(kind=4)    :: PI
    parameter(PI=3.14159265359)
    nt=size(ctmp)
    nw1=pw1*(nt/2+1)+1
    nw2=pw2*(nt/2+1)+1
    nw3=pw3*(nt/2+1)
    nw4=pw4*(nt/2+1)
    do iw=1, nt/2+1
       if(iw.ge.nw1.and.iw.le.nw2)then
          s=0.54+0.46*cos(PI*(iw-nw1)/(nw2-nw1)-PI)
          ctmp(iw)=ctmp(iw)*s
       else if(iw.ge.nw3.and.iw.le.nw4)then
          s=0.54+0.46*cos(PI*(nw3-iw)/(nw4-nw3))
          ctmp(iw)=ctmp(iw)*s
       else if(iw.gt.nw4.or.iw.lt.nw1)then
          ctmp(iw)=0.0
       end if
    end do
    do iw=nt/2+1, nt
	ctmp(iw)=ctmp(nt-iw+1)
    end do
  end subroutine

 !----------------------------------------------
  ! Hamming window taper
  !----------------------------------------------
  subroutine hammingw1(ctmp, pw3, pw4)
    complex, dimension(:)         :: ctmp
    real            :: pw3, pw4
    integer         :: nw3, nw4
    integer         :: iw, nt
    real, allocatable:: s(:)
    real(kind=4)    :: PI
    parameter(PI=3.14159265359)
    nt=size(ctmp)
    allocate(s(nt))
    s=1.0
    nw3=pw3*(nt/2+1)
    nw4=pw4*(nt/2+1)
    do iw=1, nt/2+1
       if(iw.ge.nw3.and.iw.le.nw4)then
          s(iw)=0.54+0.46*cos(PI*(nw3-iw)/(nw4-nw3))
	  ctmp(iw)=s(iw)*ctmp(iw)
       else if(iw.gt.nw4)then
	  s(iw)=0.0
	  ctmp(iw)=s(iw)*ctmp(iw)
       end if
    end do
    do iw=nt/2+1, nt	
	s(iw)=s(nt-iw+1)
	ctmp(iw)=s(iw)*ctmp(iw)
    end do
    deallocate(s)
  end subroutine

end module
