module qpertub_mod
  use prop_base
  use  base_fftwv3_mod

  implicit none

  integer, private                    :: n1, n2, nt, it
  real, private                       :: idt, w0, d1, d2
  real, dimension(:), private, pointer	      :: gama_mean 
  real, dimension(:,:), pointer, private      :: gama, tau, eta, vel, bq

  contains

  subroutine qpertub_init(n1_in, n2_in, nt_in, dt_in,d1_in, d2_in, gama_in, gama_mean_in, tau_in, eta_in, vel_in, bq_in, w0_in) 
    integer                      :: n1_in, n2_in, nt_in
    real                         :: dt_in, d1_in, d2_in
    real			:: w0_in
    real, dimension(:),target     :: gama_mean_in !!! Yi's change
    real, dimension(:,:), target   :: gama_in, tau_in, eta_in, bq_in, vel_in !!! Yi's change

    n1 = n1_in
    n2 = n2_in
    nt = nt_in
    d1= d1_in
    d2=d2_in
    idt = 1./dt_in
     w0=w0_in
     gama_mean=>gama_mean_in
     vel => vel_in
     bq => bq_in
     tau => tau_in
     eta => eta_in
     gama => gama_in 
  end subroutine

  function qpertub_op(adj, add, model, data) result(stat)
    logical,intent(in) :: adj, add
    real,dimension(:)  :: model, data
    integer            :: stat
    call qpertub_op2(adj, add, model, data)
    stat=0
  end function

  subroutine qpertub_op2(adj, add, model, data)
    logical,intent(in)          :: adj, add 
    real		:: pi
    integer :: stat, ix, iz, it, a1, a2
    real, dimension(n1,n2,nt) :: model, data
    complex, dimension(n1,n2,nt) ::chmod,clmod
    integer*8       :: plan_clmodI, plan_chmodI, plan_LmodI, plan_DLmodI, plan_HmodI, plan_DHmodI
    integer*8       :: plan_clmodF, plan_chmodF, plan_LmodF, plan_DLmodF, plan_HmodF, plan_DHmodF
    real, dimension(n1):: kz 
    real, dimension(n2):: kx 
    real, dimension(n1, n2):: k_img, oper_H, oper_L, oper_D,oper_DL, oper_DH, dgama
    complex, dimension(n1, n2)::Lmod, DLmod, Hmod, DHmod

  !!! Yi's change: 
    pi=3.14129265
    call k_map(kx,n2,d2)
    call k_map(kz,n1,d1)
    do ix=1,n2
      do iz=1,n1
	 k_img(iz, ix) =sqrt(kx(ix)*kx(ix)+kz(iz)*kz(iz))
      end do
    end do
    dgama=-1/pi/(1+bq**2)
    oper_H =(k_img)**(2*gama_mean(1)+1)
    oper_L =(k_img)**(2*gama_mean(1)+2)
    oper_D =-k_img**2

  !  do ix=1,n2
   !   do iz=1,n1

	!if (oper_L(iz,ix) .lt. maxval(oper_L)*0.3) oper_L(iz, ix)=maxval(oper_L)*0.3
	!if (oper_H(iz,ix) .lt. maxval(oper_H)*0.3) oper_H(iz, ix)=maxval(oper_H)*0.3
	!if (abs(oper_D(iz,ix)) .lt. maxval(abs(oper_D))*0.3) oper_D(iz, ix)=-maxval(abs(oper_D))*0.3
      !end do
    !end do
   ! oper_DL=oper_L*log(-oper_D)
    !oper_DH=oper_H*log(-oper_D)


    do ix=1,n2
      do iz=1,n1
	if (-oper_D(iz, ix) .gt. 0.0) then
    	  oper_DL(iz, ix)=oper_L(iz, ix)*log(-oper_D(iz, ix))
    	  oper_DH(iz, ix)=oper_H(iz, ix)*log(-oper_D(iz, ix))
	else
	write(0,*) "ix, iz", ix, iz
	  oper_DL(iz, ix)=oper_L(iz, ix)*0.0!log(-max_oper_D)
	  oper_DH(iz, ix)=oper_H(iz, ix)*0.0!log(-max_oper_D)
	end if
      end do
    end do

   
    call sfftw_init_threads(stat);
    call sfftw_plan_with_nthreads(16);

    if(adj) then
  if(.not. add) model = 0. 
  do it=1, nt
	Lmod=(2*eta*log(vel)-2*eta*log(w0)-pi*vel*tau)*dgama*data(:,:,it)
	DLmod=eta*dgama*data(:,:,it)
	Hmod=(2*tau*log(vel)-2*tau*log(w0)+pi/vel*eta)*dgama*data(:,:,it)
	DHmod=tau*dgama*data(:,:,it)

	call sfftw_dft_2dx_init(plan_clmodI, clmod(:,:,it), n1, n2, -1)
	call sfftw_dft_2dx_init(plan_chmodI, chmod(:,:,it), n1, n2, -1)
    	call sfftw_dft_2dx_init(plan_LmodF, Lmod, n1, n2, 1)
    	call sfftw_dft_2dx_init(plan_DLmodF, DLmod, n1, n2, 1)
    	call sfftw_dft_2dx_init(plan_HmodF, Hmod, n1, n2, 1)
    	call sfftw_dft_2dx_init(plan_DHmodF, DHmod, n1, n2, 1)

	call sfftw_dft_2dx_oper(plan_LmodF, Lmod, n1, n2)
	call sfftw_dft_2dx_oper(plan_DLmodF, DLmod, n1, n2)
	call sfftw_dft_2dx_oper(plan_HmodF, Hmod, n1, n2)
	call sfftw_dft_2dx_oper(plan_DHmodF, DHmod, n1, n2)
	
	clmod(:,:,it)=oper_L*Lmod+oper_DL*DLmod
	chmod(:,:,it)=oper_H*Hmod+oper_DH*DHmod

	call sfftw_dft_2dx_oper(plan_clmodI, clmod(:,:,it), n1, n2)
	call sfftw_dft_2dx_oper(plan_chmodI, chmod(:,:,it), n1, n2)	
       end do

       model(:,:,1:nt-1) = model(:,:,1:nt-1) + real((chmod(:,:,1:nt-1)-chmod(:,:,2:nt)))*idt+real(clmod(:,:,1:nt-1))
       model(:,:,nt) = model(:,:,nt) + real(chmod(:,:,nt))*idt+real(clmod(:,:,nt))
     

    else 
   
  if(.not. add) data = 0.
   clmod=model
   chmod(:,:,1)=chmod(:,:,1)+model(:,:,1)*idt
   chmod(:,:,2:nt) = chmod(:,:,2:nt)+(model(:,:,2:nt)-model(:,:,1:nt-1))*idt 

      do it=1, nt
	call sfftw_dft_2dx_init(plan_clmodF, clmod(:,:,it), n1, n2, 1)
	call sfftw_dft_2dx_init(plan_chmodF, chmod(:,:,it), n1, n2, 1)
    	call sfftw_dft_2dx_init(plan_LmodI, Lmod, n1, n2, -1)
    	call sfftw_dft_2dx_init(plan_DLmodI, DLmod, n1, n2, -1)
    	call sfftw_dft_2dx_init(plan_HmodI, Hmod, n1, n2, -1)
    	call sfftw_dft_2dx_init(plan_DHmodI, DHmod, n1, n2, -1)

	call sfftw_dft_2dx_oper(plan_clmodF, clmod(:,:,it), n1, n2)
	call sfftw_dft_2dx_oper(plan_chmodF, chmod(:,:,it), n1, n2)
	Lmod=clmod(:,:,it)*oper_L
	DLmod=clmod(:,:,it)*oper_DL
	Hmod=chmod(:,:,it)*oper_H
	DHmod=chmod(:,:,it)*oper_DH	

	call sfftw_dft_2dx_oper(plan_LmodI, Lmod, n1, n2)
	call sfftw_dft_2dx_oper(plan_DLmodI, DLmod, n1, n2)
	call sfftw_dft_2dx_oper(plan_HmodI, Hmod, n1, n2)
	call sfftw_dft_2dx_oper(plan_DHmodI, DHmod, n1, n2)	
       data(:,:,it)=(2*eta*log(vel)-2*eta*log(w0)-pi*vel*tau)*dgama*real(Lmod)+eta*dgama*real(DLmod)+(2*tau*log(vel)-2*tau*log(w0)+pi/vel*eta)*dgama*real(Hmod)+tau*dgama*real(DHmod)
       end do
     end if 
end subroutine
end module



